#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "my_types.h"
#include "form.h"
#include "solving_equations.h"

#define TIME_WINDOW 25.0
#define BED_STEPS 10
#define MIN_INTENS 0.1
#define A_NULL 0.0001


const double dr_proj = 0.000001; 

vect_t sum_v(vect_t *lhs, vect_t *rhs){
	vect_t result;
	int i;

	for (i = 0; i < 3; i++){
		*(&result.x + i) = *(&lhs->x + i) + *(&rhs->x + i);
	}

	return result;
}

vect_t dif_v(vect_t *lhs, vect_t *rhs){
	vect_t result;
	int i;

	for (i = 0; i < 3; i++){
		*(&result.x + i) = *(&lhs->x + i) - *(&rhs->x + i);
	}

	return result;
}

vect_t Vect(double x, double y, double z) {
	vect_t result = {x, y, z};
	return result;
}

double intensity(emf_t EM){
	double result = 0.0;
	int i;

	// result += EM.Ex * EM.Ex;
	// result += EM.iEx * EM.iEx;

	for (i = 0; i < sizeof(EM)/sizeof(double); i++){
		result += *(&EM.Ex + i) * *(&EM.Ex + i);
	}

	return result;
}

double delta_intensity(void (*field_definition)(vect_t, double, emf_t*, config_t),
	double t, vect_t *R, vect_t *dr, config_t cfg){
	emf_t EM;
	double result;
	double sigma = t - R[0].z + cfg.zf - cfg.tau;
	double time_profile = time_form(cfg.time_form, sigma, cfg.tau);


	field_definition(sum_v(R, dr), t, &EM, cfg);
	result = intensity(EM);
	field_definition(dif_v(R, dr), t, &EM, cfg);
	result -= intensity(EM);

	return result / 2.0 / dr_proj*time_profile*time_profile ;

}

vect_t Gradient(vect_t *R, double t,
	void (*field_definition)(vect_t, double, emf_t*, config_t), config_t cfg){
	vect_t result;
	vect_t dr;

	dr = Vect(dr_proj, 0.0, 0.0);
	result.x = delta_intensity(field_definition, t, R, &dr, cfg);

	dr = Vect(0.0, dr_proj, 0.0);
	result.y = delta_intensity(field_definition, t, R, &dr, cfg);

	dr = Vect(0.0, 0.0, dr_proj);
	result.z = delta_intensity(field_definition, t, R, &dr, cfg);

	return result;
}


Force_t make_force(void (*field_definition)(vect_t, double,emf_t*, config_t),
		 particle_t p, config_t cfg) {
	Force_t result;
	// vect_t R = {p.x, p.y, p.z};
	vect_t intens_grad =  Gradient((vect_t *) &p.x, p.time, field_definition, cfg);

	result.px = -intens_grad.x/4.0;
	result.py = -intens_grad.y/4.0;  
	result.pz = -intens_grad.z/4.0;
	result.x = p.px;
	result.y = p.py;
	result.z = p.pz;

	// printf("%lf %lf %lf\n", p.x, p.y, p.z);

		
	// 
	return result;

}



particle_t sum_p(particle_t p1, particle_t p2){
	particle_t result;
	int i;

	for (i = 0; i < 7; i++){
		*(&result.px + i) = *(&p1.px + i) + *(&p2.px + i);
	}

	return result;
}

particle_t create_Runge_step(Force_t Runge, double mult){
	particle_t result;
	int i;

	for (i = 0; i < 6; i++){
		*(&result.px + i) = *(&Runge.px + i)*mult;
	}

	result.time = mult;

	return result;
}

void Runge_Kutta(Force_t *F,particle_t* el, config_t cfg,
	void (*field_definition)(vect_t, double, emf_t*, config_t)) {
    Force_t Runge[4];
    particle_t p;
    int i, j,
    	T = cfg.steps; 
    double H = cfg.dt;



    for(i = 0; i < T; i++) {

    	(el+i+1)->time = (i + 1)*H;
   
        Runge[0] = make_force(field_definition, el[i], cfg);
        F[i] = Runge[0];

        p = create_Runge_step(Runge[0], H / 2.0);
        Runge[1] = make_force(field_definition, sum_p(el[i], p), cfg);

        p = create_Runge_step(Runge[1], H / 2.0);
        Runge[2] = make_force(field_definition, sum_p(el[i], p), cfg);

        p = create_Runge_step(Runge[2], H);
        Runge[3] = make_force(field_definition, sum_p(el[i], p), cfg);

		for (j = 0; j < 6; j ++) {
			*(&((el+i+1)->px) + j) = *(&((el+i)->px) + j) + 
				H * (*(&Runge[0].px + j) + *(&Runge[1].px + j) * 2 + 
				*(&Runge[2].px + j) * 2 + *(&Runge[3].px + j)) / 6;	
		}

		// printf ("Runge[3].x = %lf\n", Runge[3].x);
	}

	return;
}


void Adams_calc(double H, Force_t *pF, particle_t* p_el, config_t cfg,
			void (*field_definition)(vect_t, double,emf_t*, config_t)) {
	int j;

	(p_el + 1)->time = (p_el)->time + H;
	pF[0] = make_force(field_definition, p_el[0], cfg);
	for (j = 0; j < 6; j ++) {
    	*(&((p_el+1)->px) + j) = *(&(p_el->px) + j) + H*( *(&(pF->px) + j)*55
    		- *(&((pF-1)->px) + j)*59 + *(&((pF-2)->px) + j)*37 - *(&((pF-3)->px) + j)*9)/24;
    }

    // printf("el.y = %lf\n", p_el[1].y);
}

void Adams(Force_t *F,particle_t* el,config_t cfg,
	void (*field_definition)(vect_t, double,emf_t*, config_t)) {
    int i,
    	T = cfg.steps; 
    double H = cfg.dt;

    cfg.steps = 3;
    Runge_Kutta( F, el, cfg, field_definition);

        //Adams' method
    for(i = 3; i < T-1; i++) {
        Adams_calc(H, (F+i), (el+i), cfg, field_definition);
        // printf ("t(Adams) = %lf\n", el[i+1].time);
	}
}


void (*initialize_method(int calc_scheme,int rank))(Force_t *F,particle_t* el,config_t cfg,
						  void (*field)(vect_t, double, emf_t*, config_t)) {
	//initializes the fields and returns pointer to the function calculating the field
	void (*calculation)(Force_t *F,particle_t* el,config_t cfg,
			    void (*field_definition)(vect_t, double, emf_t*, config_t));

	switch(calc_scheme) {
		case 0:
			calculation = Adams;
			if (rank == 0) printf("Adams method\n");
			break;
		case 1:
			calculation = Runge_Kutta;
			if (rank == 0) printf("Runge Kutta\n");
			break;
		default:
			calculation = NULL;
			if (rank == 0) printf("Unknown method = %d", calc_scheme);
			break;

	}

	return calculation;
}